import kivy
kivy.require('1.0.8')

from kivy.core.window import Window
from kivy.uix.widget import Widget
from kivy.uix.image import Image
from kivy.graphics import Fbo, Color, Rectangle
from kivy.app import App
from kivy.clock import Clock
#from kivy.graphics.instructions import Instruction
import math
import copy
import random


def getBottomLeftPointUsingCenter(centerX, centerY, oldRectangle):
    return (centerX - oldRectangle.size[0] / 2, centerY - oldRectangle.size[1] / 2)


def getCenterX(rectangle):
    return rectangle.pos[0] + rectangle.size[0] / 2


def getCenterY(rectangle):
    return rectangle.pos[1] + rectangle.size[1] / 2


def getCenterPoint(rectangle):
    return (rectangle.pos[0] + rectangle.size[0] / 2, rectangle.pos[1] + rectangle.size[1] / 2)


def getDistance1D(value1, value2):
    result = value2 - value1
    if result < 0:
        result = result * -1
    return result


def getDistance2D(x1, y1, x2, y2):
    # **2 means squared:
    return math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)

def getDistanceBetween2DPoints(point1, point2):
    # **2 means squared:
    # coordinate [0] is x, coordinate [1] is y
    return math.sqrt((point2[0] - point1[0]) ** 2 + (point2[1] - point1[1]) ** 2)


def changeTextureCoordinatesToSpriteSheetCell(objectScreenRect, sheetWidth, sheetHeight, sheetRowsCount, sheetColumnsCount, chooseFrameNumber):
    #default tex_coords are [u, v, u + w, v, u + w, y + h, u, y + h]
    #tex_coords (for a Canvas or Rectangle) must be a list of 8 points (dest rect then source rect)
    #changing coords changes how texture (image) is mapped onto rectangle, for example: objectScreenRect.tex_coords[6]=80
    print objectScreenRect.tex_coords


#NOTE: PPF is a speed, and stands for Pixels Per Frame

pixelsPerMeter = 64


def getMetersFromPixels(length):
    global pixelsPerMeter
    return length/pixelsPerMeter


def calculateScreenRectUsingWorldRects(objectScreenRect, objectWorldRect, viewerWorldRect):
    objectScreenRect.pos = (objectWorldRect.pos[0] - viewerWorldRect.pos[0], objectWorldRect.pos[1] - viewerWorldRect.pos[1])


def getScreenRectUsingWorldRects(objectWorldRect, viewerWorldRect, sourceImageFileName):
    newRect=None
    if (sourceImageFileName is not None):
        newRect = Rectangle(pos=(objectWorldRect.pos[0],objectWorldRect.pos[1]), size=(objectWorldRect.size[0],objectWorldRect.size[1]), source=sourceImageFileName)
    else:
        newRect = copy.deepcopy(objectWorldRect)

    calculateScreenRectUsingWorldRects(newRect, objectWorldRect, viewerWorldRect)
    return newRect


def calculateWorldPointUsingScreenPoint(screenPoint, viewerWorldRect):
    return (screenPoint[0] + viewerWorldRect.pos[0], screenPoint[1] + viewerWorldRect.pos[1])



class Viewer:
    worldRect = None
    #NOTE: does not need a screen rect, since it is the screen
    def __init__(self):
        self.worldRect = Rectangle(pos=(0, 0), size=(Window.size[0], Window.size[1]))


class Entity:
    color = None
    #NOTE: if Rectangle here instead of None, screenRect will be shared by all!
    screenRect = None
    worldRect = None
    hasAI = None
    #visibility is not yet implemented:
    visible = None
    targetPoint = None
    maxSpeedPPF = None
    imageFileName = None
    teamIndex = None
    invincible = False
    hitRadiusPixelCount = None
    #disintegration is not yet implemented (for drawing explosions and other particle entities):
    disintegration = None
    isAlive = None
    canvasIndex = None
    damagesOnTouch = None

    def __init__(self):
        self.color = Color(1, 1, 1, 1)
        self.screenRect = Rectangle(pos=(0, 0), size=(64, 64))
        self.worldRect = Rectangle(pos=(0, 0), size=(64, 64))
        self.hasAI = False
        self.visible = True
        self.targetPoint = (0,0)
        self.maxSpeedPPF = 10
        self.hitRadiusPixelCount = 32
        self.disintegration = 0.0
        self.isAlive = True
        self.damagesOnTouch = False

    def refresh(self):
        incrementX = 0
        incrementY = 0
        targetX, targetY = self.targetPoint
        if (targetX < getCenterX(self.worldRect)):
            incrementX = min(self.maxSpeedPPF, getDistance1D(targetX, getCenterX(self.worldRect))) * -1
        elif (targetX > getCenterX(self.worldRect)):
            incrementX = min(self.maxSpeedPPF, getDistance1D(targetX, getCenterX(self.worldRect)))

        if targetY < getCenterY(self.worldRect):
            incrementY = min(self.maxSpeedPPF, getDistance1D(targetY, getCenterY(self.worldRect))) * -1
        elif targetY > getCenterY(self.worldRect):
            incrementY = min(self.maxSpeedPPF, getDistance1D(targetY, getCenterY(self.worldRect)))

        self.worldRect.pos = (self.worldRect.pos[0] + incrementX, self.worldRect.pos[1] + incrementY)
        #line above does same as:
        #self.worldRect.pos = getBottomLeftPointUsingCenter(getCenterX(self.worldRect) + incrementX, getCenterY(self.worldRect) + incrementY)

    def stop(self):
        #stop by setting targetPoint to location of self:
        self.targetPoint = getCenterPoint(self.worldRect)


class CanvasForm(Widget):
    entities = list()
    player1Entity = None
    player1EntityIndex = None
    currentWorld = None
    viewFollowsEntity = None
    viewFollowsEntityIndex = None
    currentViewer = None
    player1TeamIndex = 0
    lastUsedCanvasIndex = -1

    def __init__(self, **kwargs):
        super(CanvasForm, self).__init__(**kwargs)
        self._keyboard = Window.request_keyboard(
            self._keyboard_closed, self)
        self._keyboard.bind(on_key_down=self._on_keyboard_down)
        self.bind(on_touch_down=self.canvasTouchDown)

        self.currentViewer = Viewer()

        self.currentWorld = Entity()
        self.currentWorld.color = Color(1, 1, 1, 1)  # be careful, this tints the polygon even if has a texture (image)
        #Color is Red, Green, Blue, Alpha (opacity), so Color(.5,.5,.5,1) would be gray
        self.currentWorld.worldRect = Rectangle(pos=(0, 0), size=(2000, 2000))
        self.currentWorld.screenRect = Rectangle(source="area1.png", pos=(0, 0), size=(2000, 2000))

        self.canvas.add(self.currentWorld.color)
        self.canvas.add(self.currentWorld.screenRect)
        self.lastUsedCanvasIndex = self.lastUsedCanvasIndex + 1
        #NOTES:
        #pos[0] is x coordinate for position
        #pos[1] is y coordinate for position
        #The color constructor takes Red,Green,Blue, and Alpha which is opacity
        #so Color(1,0,0,1) is red

        #This is player 1:
        lastCreatedEntity = Entity()
        self.entities.append(lastCreatedEntity)
        lastCreatedEntityIndex = len(self.entities) - 1
        self.player1Entity = lastCreatedEntity
        self.player1EntityIndex = lastCreatedEntityIndex
        self.viewFollowsEntityIndex = lastCreatedEntityIndex
        self.viewFollowsEntity = lastCreatedEntity

        lastCreatedEntity.color = Color(1, 1, 1, 1)
        lastCreatedEntity.screenRect = Rectangle(source="Car, Rear - lincoln_tc_fournel (towncar, cholasimmons at turbosquid) - MrG\'s Render1.png", pos=lastCreatedEntity.worldRect.pos, size=lastCreatedEntity.worldRect.size)
        player1StartX = self.currentWorld.worldRect.size[0] / 2
        player1StartY = self.currentWorld.worldRect.size[1] / 2
        lastCreatedEntity.worldRect.pos = getBottomLeftPointUsingCenter(player1StartX, player1StartY, self.entities[0].worldRect)
        self.canvas.add(lastCreatedEntity.color)
        self.canvas.add(lastCreatedEntity.screenRect)
        self.lastUsedCanvasIndex = self.lastUsedCanvasIndex + 1
        lastCreatedEntity.canvasIndex = self.lastUsedCanvasIndex
        lastCreatedEntity.stop()
        lastCreatedEntity.teamIndex = self.player1TeamIndex

        #This is the enemy:
        lastCreatedEntity = Entity()
        self.entities.append(lastCreatedEntity)
        lastCreatedEntityIndex = len(self.entities) - 1

        lastCreatedEntity.color = Color(1, 1, 1, 1)
        lastCreatedEntity.screenRect = Rectangle(source="Rock 1 by BesideTheVoid - edited from IMG_3319 by victorblagovici on morguefile.png", pos=lastCreatedEntity.worldRect.pos, size=lastCreatedEntity.worldRect.size)
        self.canvas.add(lastCreatedEntity.color)
        self.canvas.add(lastCreatedEntity.screenRect)
        self.lastUsedCanvasIndex = self.lastUsedCanvasIndex + 1
        lastCreatedEntity.canvasIndex = self.lastUsedCanvasIndex
        enemyStartX = random.randrange(0, self.currentWorld.worldRect.size[0])
        enemyStartY = random.randrange(0, self.currentWorld.worldRect.size[1])
        lastCreatedEntity.worldRect.pos = (enemyStartX, enemyStartY)
        lastCreatedEntity.hasAI = True
        #Make the enemy move toward the player:
        lastCreatedEntity.targetPoint = getCenterPoint(self.player1Entity.worldRect)
        lastCreatedEntity.teamIndex = 1
        lastCreatedEntity.damagesOnTouch = True

        #Making update interval 1/60 sec makes game run at 60 frames per second:
        Clock.schedule_interval(self.update, 1.0 / 60.0)

    def getWidth(self):
        return self.size[0]

    def getHeight(self):
        return self.size[1]

    def explodeEntity(self, thisEntity):
        if (thisEntity.invincible is False):
            thisEntity.isAlive = False
        #thisEntity.screenRect.visible = False
        #self.canvas.remove("player1Instruction")
        thisEntity.maxSpeedPPF = 0
        #Setting a (alpha) to zero makes image invisible, but don't do this if need explosion on that layer of canvas
        thisEntity.color.a = 0
        self.addExplosion(thisEntity.worldRect, 3, thisEntity.canvasIndex)

    def addExplosion(self, destinationWorldRect, scale, canvasIndex):
        self.entities.append(Entity())
        #lastCreatedEntityIndex = len(self.entities) - 1
        lastCreatedEntity = Entity()
        #lastCreatedEntity.worldRect = copy.deepcopy(destinationWorldRect)
        scaledSize = (destinationWorldRect.size[0] * scale, destinationWorldRect.size[1] * scale)
        lastCreatedEntity.worldRect = Rectangle(pos=(destinationWorldRect.pos[0], destinationWorldRect.pos[1]), size = scaledSize)
        lastCreatedEntity.worldRect.pos = getBottomLeftPointUsingCenter(getCenterX(destinationWorldRect), getCenterY(destinationWorldRect), lastCreatedEntity.worldRect)
        #lastCreatedEntity.worldRect.pos = (lastCreatedEntity.worldRect.pos[0]+64,lastCreatedEntity.worldRect.pos[1]+64)
        #lastCreatedEntity.worldRect.size = (lastCreatedEntity.worldRect.size[0]*scale, lastCreatedEntity.worldRect.size[1]*scale)
        lastCreatedEntity.screenRect = Rectangle(source="Sprite_Explosion_of_sorts_by_leileilol (BesideTheVoid plain opacity version) frame0012.png", pos=(lastCreatedEntity.worldRect.pos[0], lastCreatedEntity.worldRect.pos[1]), size=(lastCreatedEntity.worldRect.size[0], lastCreatedEntity.worldRect.size[1]))
        #lastCreatedEntity.screenRect = Rectangle(source="Sprite_Explosion_of_sorts_by_leileilol (BesideTheVoid SPRITE SHEET plain opacity version).gif", pos=(destinationWorldRect.pos[0],destinationWorldRect.pos[1]), size=(destinationWorldRect.size[0],destinationWorldRect.size[1]))

        #changeTextureCoordinatesToSpriteSheetCell(lastCreatedEntity.screenRect, 512, 512, 4, 4, 11)
        #lastCreatedEntity.screenRect = getScreenRectUsingWorldRects(destinationWorldRect, self.currentViewer.worldRect, "Sprite_Explosion_of_sorts_by_leileilol (BesideTheVoid SPRITE SHEET plain opacity version).gif")
        calculateScreenRectUsingWorldRects(lastCreatedEntity.screenRect, lastCreatedEntity.worldRect, self.currentViewer.worldRect)
        lastCreatedEntity.invincible = True
        #TODO: find out why insert doesn't work:
        #if ((canvasIndex is not None) and (canvasIndex>=0)):
            #self.canvas.insert(canvasIndex,Color(1,1,1,1))
            #self.canvas.insert(canvasIndex,lastCreatedEntity.screenRect)
        #else:
        self.canvas.add(Color(1, 1, 1, 1))
        self.canvas.add(lastCreatedEntity.screenRect)

    def _keyboard_closed(self):
        print 'Keyboard disconnectd!'
        self._keyboard.unbind(on_key_down=self._on_keyboard_down)
        self._keyboard = None

    def update(self, dt):
        for index in range(0, len(self.entities)):
            thisEntity = self.entities[index]
            self.entities[index].refresh()
            if (index == self.viewFollowsEntityIndex):
                #must update viewer after the entity being tracked moves (it actually be moved first):
                self.currentViewer.worldRect.pos = getBottomLeftPointUsingCenter(getCenterX(self.viewFollowsEntity.worldRect), getCenterY(self.viewFollowsEntity.worldRect), self.currentViewer.worldRect)
            calculateScreenRectUsingWorldRects(thisEntity.screenRect, thisEntity.worldRect, self.currentViewer.worldRect)

        calculateScreenRectUsingWorldRects(self.currentWorld.screenRect, self.currentWorld.worldRect, self.currentViewer.worldRect)

    def _on_keyboard_down(self, keyboard, keycode, text, modifiers):
        print 'The key', keycode, ' pressed'
        print ' - text is %r' % text
        print ' - modifiers are %r' % modifiers
        # Keycode is composed of an integer + a string
        # If we hit escape, release the keyboard
        if keycode[1] == 'escape':
            keyboard.release()
        return True

    def canvasTouchDown(self, touch, *largs):
        #touchX, touchY = largs[0].pos
        self.player1Entity.targetPoint = calculateWorldPointUsingScreenPoint(largs[0].pos, self.currentViewer.worldRect)


class HardModeGPSApp(App):
    def build(self):
        return CanvasForm()

if __name__ == '__main__':
    HardModeGPSApp().run()
